from typing import Generator, List
import regex as re
import string
from pie.tagger import regexsplitter, SECTION, FULLSTOP

WORD = r'([{}])'.format(string.punctuation)


class SimpleTokenizer(object):
    """ Tokenizer that memoryze what it tokenized.

    Mostly used to normalized input as input time and then reinserting normalized input

    """
    def __init__(self):
        self.section = regexsplitter(SECTION)
        self.fullstop = regexsplitter(FULLSTOP)
        self.word = regexsplitter(WORD)

    def sentence_tokenizer(self, text: str, lower: bool = False) -> Generator[List[str], None, None]:
        for line in self.section(text):
            for sentence in self.fullstop(line):
                yield self.word_tokenizer(sentence, lower=lower)

    def word_tokenizer(self, text: str, lower: bool = False) -> List[str]:
        sentence = [w for raw in text.split() for w in self.word(raw)]
        if lower:
            sentence = [w.lower() for w in sentence]
        return sentence

    def reset(self):
        """Can be used between documents for example """
        pass



import regex as re
from typing import List, Generator

from pie_extended.pipeline.tokenizers.memorizing import MemorizingTokenizer

_Dots_except_apostrophe = r".?!\"“”\"«»…\[\]\(\)„“"
_Dots_collections = r"[" + _Dots_except_apostrophe + "‘’]"
_RomanNumber = r"(?:M{1,4}(?:CM|CD|D?C{0,3})(?:XC|XL|L?X{0,3})" \
               r"(?:IX|IV|V?I{0,3})|M{0,4}(?:CM|C?D|D?C{1,3})" \
               r"(?:XC|XL|L?X{0,3})(?:IX|IV|V?I{0,3})|M{0,4}" \
               r"(?:CM|CD|D?C{0,3})(?:XC|X?L|L?X{1,3})" \
               r"(?:IX|IV|V?I{0,3})|M{0,4}(?:CM|CD|D?C{0,3})" \
               r"(?:XC|XL|L?X{0,3})(?:IX|I?V|V?I{1,3}))"


class FroMemorizingTokenizer(MemorizingTokenizer):
    APOSTROPHES = "'’ʼ"
    re_elision_apostrophe = re.compile(r"(\w+)([" + APOSTROPHES + r"])(\w+)")
    re_add_space_around_punct = re.compile(r"(\s*)([^\w\s])(\s*)")
    re_add_space_around_apostrophe_that_are_quotes = re.compile(
        r"("
        r"(((?<=[\W])[\'’ʼ]+(?=[\W]))|"
        r"((?<=[\w])[\'’ʼ]+(?=[\W]))|"
        r"((?<=[\W])[\'’ʼ]+(?=[\w])))|"
        r"(^[\'’ʼ]+)|"
        r"([\'’ʼ]+$))"
        # NotLetter+Apo+NotLetter or Letter+Apo+NotLetter or NotLetter+Apo+Letter + Starting or ending apostrophe
        # ?'. or manger'_ or _'Bonjour
    )
    re_add_space_after_apostrophe = re.compile(r"(\s*)([\'’ʼ])(\s*)")
    re_remove_ending_apostrophe = re.compile(r"(?<=\w)([\'’ʼ])")
    _sentence_boundaries = re.compile(
        r"([" + _Dots_except_apostrophe + r"]+\s*)+"
    )
    roman_number_dot = re.compile(r"\.(" + _RomanNumber + r")\.")

    def __init__(self):
        super(FroMemorizingTokenizer, self).__init__()
        self.tokens = []

    @staticmethod
    def _sentence_tokenizer_merge_matches(match):
        """ Best way we found to deal with repeating groups"""
        start, end = match.span()
        return match.string[start:end] + "<SPLIT>"

    def _real_sentence_tokenizer(self, string: str) -> List[str]:
        string = self._sentence_boundaries.sub(self._sentence_tokenizer_merge_matches, string)
        string = string.replace("_DOT_", ".")
        for index_apo, apo in enumerate(self.APOSTROPHES):
            string = string.replace("ApOsTrOpHe"+str(index_apo), apo+" ")
        return string.split("<SPLIT>")

    def _real_word_tokenizer(self, text: str, lower: bool = False) -> List[str]:
        if lower:
            text = text.lower()
        text = text.split()
        return text

    def sentence_tokenizer(self, text: str, lower: bool = False) -> Generator[List[str], None, None]:
        sentences = list()
        data = self.normalizer(text)
        for sent in self._real_sentence_tokenizer(data):
            sent = sent.strip()
            if sent:
                sentences.append(self.word_tokenizer(sent))
        yield from sentences

    def apostrophe_replace(self, regex_match) -> str:
        return regex_match.group(1) + "ApOsTrOpHe"+ str(self.APOSTROPHES.index(regex_match.group(2))) + regex_match.group(3)

    def normalizer(self, data: str) -> str:
        data = self.re_add_space_around_punct.sub(
                    r" \g<2> ",
                    self.re_elision_apostrophe.sub(
                        self.apostrophe_replace,
                        self.roman_number_dot.sub(
                            r"_DOT_\g<1>_DOT_",
                            data
                        )
                    )
        )
        return data

    def replacer(self, inp: str):
        return self.re_remove_ending_apostrophe.sub("", inp)



import regex as re
from typing import List, Generator

from pie_extended.pipeline.tokenizers.memorizing import MemorizingTokenizer

_Dots_except_apostrophe = r".?!\"“”\"«»…\[\]\(\)„“"
_Dots_collections = r"[" + _Dots_except_apostrophe + "‘’]"
_RomanNumber = r"(?:M{1,4}(?:CM|CD|D?C{0,3})(?:XC|XL|L?X{0,3})" \
               r"(?:IX|IV|V?I{0,3})|M{0,4}(?:CM|C?D|D?C{1,3})" \
               r"(?:XC|XL|L?X{0,3})(?:IX|IV|V?I{0,3})|M{0,4}" \
               r"(?:CM|CD|D?C{0,3})(?:XC|X?L|L?X{1,3})" \
               r"(?:IX|IV|V?I{0,3})|M{0,4}(?:CM|CD|D?C{0,3})" \
               r"(?:XC|XL|L?X{0,3})(?:IX|I?V|V?I{1,3}))"


class FrMemorizingTokenizer(MemorizingTokenizer):
    APOSTROPHES = "'’ʼ"
    re_elision_apostrophe = re.compile(r"(\w+)([" + APOSTROPHES + r"])(\w+)")
    re_aujourdhui = re.compile("(aujourd)(["+APOSTROPHES+"])(hui)", flags=re.IGNORECASE)
    re_apostrophe = re.compile("(字1)")
    re_add_space_around_punct = re.compile(r"(\s*)([^\w\s])(\s*)")
    re_add_space_around_apostrophe_that_are_quotes = re.compile(
        r"("
        r"(((?<=[\W])[\'’ʼ]+(?=[\W]))|"
        r"((?<=[\w])[\'’ʼ]+(?=[\W]))|"
        r"((?<=[\W])[\'’ʼ]+(?=[\w])))|"
        r"(^[\'’ʼ]+)|"
        r"([\'’ʼ]+$))"
        # NotLetter+Apo+NotLetter or Letter+Apo+NotLetter or NotLetter+Apo+Letter + Starting or ending apostrophe
        # ?'. or manger'_ or _'Bonjour
    )
    re_add_space_after_apostrophe = re.compile(r"(\s*)([\'’ʼ])(\s*)")
    re_remove_ending_apostrophe = re.compile(r"(?<=\w)([\'’ʼ])")
    _sentence_boundaries = re.compile(
        r"([" + _Dots_except_apostrophe + r"]+\s*)+"
    )
    roman_number_dot = re.compile(r"\.(" + _RomanNumber + r")\.")
    # Need to deal with (Case insensitive)
    # - Aujourd'hui
    # - ['-ce', '-ci', '-elle', '-elles', '-en', '-eux', '-il', '-ils', '-je', '-la', '-le', '-les', '-leur', '-leurs',
    #       '-lui', '-là', '-m', '-me', '-moi', '-même', '-mêmes', '-nous', '-on', '-t', '-te', '-toi', '-tu', '-un',
    #       '-une', '-unes', '-uns', '-vous', '-y']
    # - peut-être, peut-estre, sur-tout, long-temps, par-tout, vis-à-vis
    # a-t-il -> a / -t-il

    _data_re_keep_clitics_pronouns = [
        '-ce', '-ci', '-elle', '-elles', '-en', '-eux', '-il', '-ils', '-je', '-la', '-le',
        '-les', '-leur', '-leurs',
        '-lui', '-là', '-m', '-me', '-moi', '-même', '-mêmes', '-nous', '-on', '-t',
        '-te', '-toi', '-tu', '-un',
        '-une', '-unes', '-uns', '-vous', '-y',
        r'-m字\d', r'-t字\d'  # 字 replaces all kinds of apostrophe
    ]

    re_keep_clitics = re.compile(
        r"(-)("+r"|".join(
            [tok.replace("-", "") for tok in sorted(_data_re_keep_clitics_pronouns, key=len, reverse=True)]
        )+r")(\b|$)",
        flags=re.IGNORECASE
    )

    # Not used currently
    _data_re_keep_together = "peut-être, peut-estre, sur-tout, long-temps, par-tout, vis-à-vis".split(", ")
    re_keep_together = re.compile(
        r"("+"|".join([
            token
            for token in _data_re_keep_together
        ])+r")(?:\b|$)",
        flags=re.IGNORECASE
    )

    # Abbreviations separated by `; ` (SEMICOLON+SPACE)
    # You can use regexp inside this list, such as [Cc]f. but I would rather recommend to add both `cf.; Cf;`
    #   for readability.
    _data_re_abbr = sorted(list(set(
        "Acad.; Adj.; Agricol.; Agricul.; Apocal.; anc.; Bot.; Botan.; Botaniq.; ca.; cap.; capi.; cf.; Cf.; " 
        "Cha.; Chap.; Col.; Dic.; Diction.; Dictionn.; Eccl.; Écon.; Élem.; Fig.; Fr.; Geog.; Gram.; " 
        "Gramm.; Hist.; Ibid.; Ibid.; Inst.; Jard.; Jurisprud.; Latit.; Li.; Lib.; Libr.; Lig.; Lit.; " 
        "Littérat.; Liv.; Long.; Mar.; Mat.; Mathém.; Mech.; Med.; Med.; Mem.; Menuis.; Milit.; Mod.; " 
        "Mor.; Mr.; Monsr.; nat.; natur.; N.b.; Orat.; Ornith.; Ornythol.; Ornitholog.; Part.; Pag.; " 
        "Pharm.; Phil.; Philos.; Pl.; Pl.; Politiq.; P.S.; Phys.; Physiq.; Sr.; St.; Subst.; s.f.; " 
        "S.M.; s.m.; Tab.; Théât.; Trév.; Tom.; Vol.; V. n.; V. a.; V. act.; Zoo.; " 
        "Zoolog.".split("; ")
    )), key=len, reverse=True)
    re_abbr = re.compile(
        r"\b("+"|".join([
            token.replace(" ", r"\s+").replace(".", r"\.")
            for token in _data_re_abbr
            if token
        ])+r")"
    )

    def __init__(self):
        super(FrMemorizingTokenizer, self).__init__()
        self.tokens = []

    @staticmethod
    def _sentence_tokenizer_merge_matches(match):
        """ Best way we found to deal with repeating groups"""
        start, end = match.span()
        return match.string[start:end] + "<SPLIT>"

    def _real_sentence_tokenizer(self, string: str) -> List[str]:
        string = self._sentence_boundaries.sub(self._sentence_tokenizer_merge_matches, string)

        for index_apo, apo in enumerate(self.APOSTROPHES):
            string = string.replace("字"+str(index_apo), apo)

        string = string.replace("界t 界", "-t-")
        string = string.replace("界", "-")  # Any dash separating pronouns
        #string = string.replace("分", "-")  # Agglutinated words such as peut-être
        string = string.replace("語", ".")  # Dots from abbreviations
        return string.split("<SPLIT>")

    def _real_word_tokenizer(self, text: str, lower: bool = False) -> List[str]:
        if lower:
            text = text.lower()
        text = text.split()
        return text

    def sentence_tokenizer(self, text: str, lower: bool = False) -> Generator[List[str], None, None]:
        sentences = list()
        data = self.normalizer(text)
        for sent in self._real_sentence_tokenizer(data):
            sent = sent.strip()
            if sent:
                sentences.append(self.word_tokenizer(sent))
        yield from sentences

    def replace_apostrophe(self, regex_match) -> str:
        return regex_match.group(1) + "字" + str(self.APOSTROPHES.index(regex_match.group(2))) \
               + " " + regex_match.group(3)

    def replace_keep_together(self, regex_match) -> str:
        return regex_match.group(0).replace("-", "分")

    def replace_aujourdhui(self, regex_match) -> str:
        return regex_match.group(1) + "字" + str(self.APOSTROPHES.index(regex_match.group(2))) + regex_match.group(3)

    def replace_abbr_dot(self, regex_match) -> str:
        return regex_match.group(1).replace(".", "語")

    def normalizer(self, data: str) -> str:
        data = self.re_add_space_around_punct.sub(
                    r" \g<2> ",
                    self.re_abbr.sub(
                        self.replace_abbr_dot,
                        # peut-etre, etc.
                        #self.re_keep_together.sub(
                        #    self.replace_keep_together,
                            self.re_keep_clitics.sub(
                                r" 界\2",
                                self.re_elision_apostrophe.sub(
                                    self.replace_apostrophe,
                                    self.re_aujourdhui.sub(
                                        self.replace_aujourdhui,
                                        data
                                    )
                                )
                            )
                        #)
                    )
                )
        return data

    def replacer(self, inp: str):
        return self.re_remove_ending_apostrophe.sub("'", inp)\
            .replace("-t-", "-")  # Temp feature until retrain has been done



from .simple_tokenizer import SimpleTokenizer
from typing import List, Tuple, Dict


class MemorizingTokenizer(SimpleTokenizer):
    """ Tokenizer that memoryze what it tokenized.

    Mostly used to normalized input as input time and then reinserting normalized input

    """

    def replacer(self, token: str) -> str:
        """ This function allows for changing input and keeping it in memory """
        return token

    def __init__(self):
        self.tokens: List[Tuple[int, str, str]] = []

    def _real_word_tokenizer(self, data: str, lower: bool = False) -> List[str]:
        return super(MemorizingTokenizer, self).word_tokenizer(data, lower=lower)

    def word_tokenizer(self, text: str, lower: bool = False) -> List[str]:
        sentence = []
        for token in self._real_word_tokenizer(text, lower):
            out = self.replacer(token)
            self.tokens.append((len(self.tokens), token, out))
            sentence.append(out)
        return sentence

    def reset(self):  # Empty
        self.tokens = []




from pie_extended.models.lasla.imports import get_iterator_and_processor
from pie_extended.models import lasla
from pie_extended.testing_utils import FakeTagger, create_auto_tagger
from typing import List, Tuple
import csv

from unittest import TestCase


def write_crazy_file() -> str:
    filename = "crazy_text_file.txt"
    with open(filename, "w") as f:
        f.write("""\\\\\\<1>[$@$](V)\\\\\\§
\\\\<1>[$@$]\\\\§
§
\\[I]\\§
At o sceleste penis, o meum malum,
graui piaque lege noxiam lues.
Licet querare, nec tibi tener puer""")
    return filename


def make_fake_data(sentences: List[str], nb_tasks: int = 9) -> List[Tuple[str, List[str]]]:
    return [
        [
            (token, tuple([token, *["fake"]*nb_tasks]))
            for token in sentence.split()
        ]
        for sentence in sentences
    ]


def make_controller(sentences: List[str]):
    # Add the lemmatizer routes
    tagger = FakeTagger(
        make_fake_data(sentences),
        tasks="lemma,Voice,Mood,Deg,Numb,Person,Tense,Case,Gend,pos".split(",")
    )
    iterator, processor = get_iterator_and_processor()
    return tagger, iterator, processor


class TestLasla(TestCase):
    def test_consecutive_dots(self):
        """Check that consecutive punctation does not break anything

        Found out the hard way it would break things
        """

        tagger, data_iterator, processor = make_controller([
            "id enim ait turbabuntur a facie eius patris or phanorum et iudicis uiduarum",
            "causam turbationis hanc docuit quod pater"
        ])

        result = tagger.tag_str(
            data="id enim ait turbabuntur a facie eius patris or phanorum et iudicis uiduarum ."
                          "  .  causam turbationis hanc docuit quod pater",
            processor=processor,
            iterator=data_iterator
        )
        self.assertEqual(
            result[12],
            {"form": "uiduarum", "lemma": "uiduarum", "pos": "fake", "morph": "Case=fake|Numb=fake|Deg=fake|Mood=fake|"
                                                                            "Tense=fake|Voice=fake|Person=fake",
             "treated": "uiduarum"},
            "Punctuation should be reinserted and mostly should not break anything"
        )

    def test_leading_punctuation(self):
        """Check that consecutive punctation does not break anything

        Special case of consecutive dots, where sentences starts with it
        """
        tagger, data_iterator, processor = make_controller([
            # Need an empty sentence because ( was treated as such
            "id enim ait", "turbabuntur a facie eius patris or phanorum et iudicis uiduarum"
        ])
        result = tagger.tag_str(
            "( id enim ait) turbabuntur a facie eius patris or phanorum et iudicis uiduarum ..",
            processor=processor,
            iterator=data_iterator
        )
        tokens = [t["form"] for t in result]
        self.assertEqual(
            ["(", "id", "enim", "ait", ")", "turbabuntur", "a", "facie", "eius", "patris", "or", "phanorum",
             "et", "iudicis", "uiduarum", ".", "."],
            tokens,
            "Leading punctuation should not break anything"
        )

    def test_punctuation_is_not_seen(self):
        """Check that punctuation is not seen by the tagger

        """
        tagger, data_iterator, processor = make_controller([
            "id enim ait", "turbabuntur a facie eius patris or phanorum et iudicis uiduarum"
        ])
        tagger.tag_str(
            "( id enim ait ) turbabuntur a facie eius patris or phanorum et iudicis uiduarum .  .",
            processor=processor,
            iterator=data_iterator
        )
        self.assertNotIn(
            "(",
            [tok for sent in tagger.seen for tok in sent],
            "Punctuation should not be seen by the Tagger"
        )

    def test_j_are_temporarly_replaced(self):
        """Check that characters are replaced for the tagger, thus avoiding out of domain, and reinserted

        """
        tagger, data_iterator, processor = make_controller([
            "iudicis uiduarum"
        ])
        result = tagger.tag_str(
            "judicis uiduarum",
            processor=processor,
            iterator=data_iterator
        )
        flatten_seen = list([tok for sent in tagger.seen for tok in sent])

        self.assertEqual(result[0]["form"], "judicis", "'j' should be removed from tagging")
        self.assertEqual(result[0]["treated"], "iudicis", "And 'i' should replace it")

    def test_underscores(self):
        string = "una operatio in ecclesiae fundamento.._... _ . laetatur autem pater quia filius perierat"
        tagger, data_iterator, processor = make_controller([
            "una operatio in ecclesiae fundamento", "laetatur autem pater quia filius perierat"
        ])
        tagger.tag_str(
            string,
            processor=processor,
            iterator=data_iterator
        )
        flatten_seen = list([tok for sent in tagger.seen for tok in sent])
        self.assertEqual(
            ['una', 'operatio', 'in', 'ecclesiae', 'fundamento', 'laetatur', 'autem', 'pater', 'quia', 'filius',
             'perierat'],
            flatten_seen,
            "Seen element should not count the underscord"
        )

    def test_with_fake_advanced_tagger(self):
        target = write_crazy_file()
        tagger, it, pr = create_auto_tagger(lasla, lower=True)
        out_file = tagger.tag_file(target, it, pr)
        content = []
        with open(out_file) as f:
            header = []
            for line in f:
                splitted = line.strip().split()
                if not header:
                    header = splitted
                    continue
                content.append(dict(list(zip(header, splitted))))
        self.maxDiff = None
        self.assertEqual(
                content,
            [{'token': '\\', 'lemma': '\\', 'pos': 'PUNC', 'morph': 'MORPH=empty', 'treated': '--IGN.--'},
             {'token': '\\', 'lemma': '\\', 'pos': 'PUNC', 'morph': 'MORPH=empty', 'treated': '--IGN.--'},
             {'token': '\\', 'lemma': '\\', 'pos': 'PUNC', 'morph': 'MORPH=empty', 'treated': '--IGN.--'},
             {'token': '<', 'lemma': '<', 'pos': 'PUNC', 'morph': 'MORPH=empty', 'treated': '--IGN.--'},
             {'token': '1', 'lemma': 'lemma0', 'pos': 'pos0',
              'morph': 'Case=Case0|Numb=Numb0|Deg=Deg0|Mood=Mood0|Tense=Tense0|Voice=Voice0|Person=Person0',
              'treated': '1'},
             {'token': '>', 'lemma': '>', 'pos': 'PUNC', 'morph': 'MORPH=empty', 'treated': '--IGN.--'},
             {'token': '[', 'lemma': '[', 'pos': 'PUNC', 'morph': 'MORPH=empty', 'treated': '--IGN.--'},
             {'token': '$', 'lemma': '$', 'pos': 'PUNC', 'morph': 'MORPH=empty', 'treated': '--IGN.--'},
             {'token': '@', 'lemma': '@', 'pos': 'PUNC', 'morph': 'MORPH=empty', 'treated': '--IGN.--'},
             {'token': '$', 'lemma': '$', 'pos': 'PUNC', 'morph': 'MORPH=empty', 'treated': '--IGN.--'},
             {'token': ']', 'lemma': ']', 'pos': 'PUNC', 'morph': 'MORPH=empty', 'treated': '--IGN.--'},
             {'token': '(', 'lemma': '(', 'pos': 'PUNC', 'morph': 'MORPH=empty', 'treated': '--IGN.--'},
             {'token': 'V', 'lemma': 'lemma1', 'pos': 'pos1',
              'morph': 'Case=Case1|Numb=Numb1|Deg=Deg1|Mood=Mood1|Tense=Tense1|Voice=Voice1|Person=Person1',
              'treated': '3'},
             {'token': ')', 'lemma': ')', 'pos': 'PUNC', 'morph': 'MORPH=empty', 'treated': '--IGN.--'},
             {'token': '\\', 'lemma': '\\', 'pos': 'PUNC', 'morph': 'MORPH=empty', 'treated': '--IGN.--'},
             {'token': '\\', 'lemma': '\\', 'pos': 'PUNC', 'morph': 'MORPH=empty', 'treated': '--IGN.--'},
             {'token': '\\', 'lemma': '\\', 'pos': 'PUNC', 'morph': 'MORPH=empty', 'treated': '--IGN.--'},
             {'token': '§', 'lemma': '§', 'pos': 'PUNC', 'morph': 'MORPH=empty', 'treated': '--IGN.--'},
             {'token': '\\', 'lemma': '\\', 'pos': 'PUNC', 'morph': 'MORPH=empty', 'treated': '--IGN.--'},
             {'token': '\\', 'lemma': '\\', 'pos': 'PUNC', 'morph': 'MORPH=empty', 'treated': '--IGN.--'},
             {'token': '<', 'lemma': '<', 'pos': 'PUNC', 'morph': 'MORPH=empty', 'treated': '--IGN.--'},
             {'token': '1', 'lemma': 'lemma2', 'pos': 'pos2',
              'morph': 'Case=Case2|Numb=Numb2|Deg=Deg2|Mood=Mood2|Tense=Tense2|Voice=Voice2|Person=Person2',
              'treated': '1'},
             {'token': '>', 'lemma': '>', 'pos': 'PUNC', 'morph': 'MORPH=empty', 'treated': '--IGN.--'},
             {'token': '[', 'lemma': '[', 'pos': 'PUNC', 'morph': 'MORPH=empty', 'treated': '--IGN.--'},
             {'token': '$', 'lemma': '$', 'pos': 'PUNC', 'morph': 'MORPH=empty', 'treated': '--IGN.--'},
             {'token': '@', 'lemma': '@', 'pos': 'PUNC', 'morph': 'MORPH=empty', 'treated': '--IGN.--'},
             {'token': '$', 'lemma': '$', 'pos': 'PUNC', 'morph': 'MORPH=empty', 'treated': '--IGN.--'},
             {'token': ']', 'lemma': ']', 'pos': 'PUNC', 'morph': 'MORPH=empty', 'treated': '--IGN.--'},
             {'token': '\\', 'lemma': '\\', 'pos': 'PUNC', 'morph': 'MORPH=empty', 'treated': '--IGN.--'},
             {'token': '\\', 'lemma': '\\', 'pos': 'PUNC', 'morph': 'MORPH=empty', 'treated': '--IGN.--'},
             {'token': '§', 'lemma': '§', 'pos': 'PUNC', 'morph': 'MORPH=empty', 'treated': '--IGN.--'},
             {'token': '§', 'lemma': '§', 'pos': 'PUNC', 'morph': 'MORPH=empty', 'treated': '--IGN.--'},
             {'token': '\\', 'lemma': '\\', 'pos': 'PUNC', 'morph': 'MORPH=empty', 'treated': '--IGN.--'},
             {'token': '[', 'lemma': '[', 'pos': 'PUNC', 'morph': 'MORPH=empty', 'treated': '--IGN.--'},
             {'token': 'I', 'lemma': 'lemma3', 'pos': 'pos3',
              'morph': 'Case=Case3|Numb=Numb3|Deg=Deg3|Mood=Mood3|Tense=Tense3|Voice=Voice3|Person=Person3',
              'treated': '1'},
             {'token': ']', 'lemma': ']', 'pos': 'PUNC', 'morph': 'MORPH=empty', 'treated': '--IGN.--'},
             {'token': '\\', 'lemma': '\\', 'pos': 'PUNC', 'morph': 'MORPH=empty', 'treated': '--IGN.--'},
             {'token': '§', 'lemma': '§', 'pos': 'PUNC', 'morph': 'MORPH=empty', 'treated': '--IGN.--'},
             {'token': 'At', 'lemma': 'lemma4', 'pos': 'pos4',
              'morph': 'Case=Case4|Numb=Numb4|Deg=Deg4|Mood=Mood4|Tense=Tense4|Voice=Voice4|Person=Person4',
              'treated': 'at'},
             {'token': 'o', 'lemma': 'lemma5', 'pos': 'pos5',
              'morph': 'Case=Case5|Numb=Numb5|Deg=Deg5|Mood=Mood5|Tense=Tense5|Voice=Voice5|Person=Person5',
              'treated': 'o'},
             {'token': 'sceleste', 'lemma': 'lemma6', 'pos': 'pos6',
              'morph': 'Case=Case6|Numb=Numb6|Deg=Deg6|Mood=Mood6|Tense=Tense6|Voice=Voice6|Person=Person6',
              'treated': 'sceleste'},
             {'token': 'penis', 'lemma': 'lemma7', 'pos': 'pos7',
              'morph': 'Case=Case7|Numb=Numb7|Deg=Deg7|Mood=Mood7|Tense=Tense7|Voice=Voice7|Person=Person7',
              'treated': 'penis'},
             {'token': ',', 'lemma': ',', 'pos': 'PUNC', 'morph': 'MORPH=empty', 'treated': '--IGN.--'},
             {'token': 'o', 'lemma': 'lemma8', 'pos': 'pos8',
              'morph': 'Case=Case8|Numb=Numb8|Deg=Deg8|Mood=Mood8|Tense=Tense8|Voice=Voice8|Person=Person8',
              'treated': 'o'},
             {'token': 'meum', 'lemma': 'lemma9', 'pos': 'pos9',
              'morph': 'Case=Case9|Numb=Numb9|Deg=Deg9|Mood=Mood9|Tense=Tense9|Voice=Voice9|Person=Person9',
              'treated': 'meum'},
             {'token': 'malum', 'lemma': 'lemma10', 'pos': 'pos10',
              'morph': 'Case=Case10|Numb=Numb10|Deg=Deg10|Mood=Mood10|Tense=Tense10|Voice=Voice10|Person=Person10',
              'treated': 'malum'},
             {'token': ',', 'lemma': ',', 'pos': 'PUNC', 'morph': 'MORPH=empty', 'treated': '--IGN.--'},
             {'token': 'graui', 'lemma': 'lemma11', 'pos': 'pos11',
              'morph': 'Case=Case11|Numb=Numb11|Deg=Deg11|Mood=Mood11|Tense=Tense11|Voice=Voice11|Person=Person11',
              'treated': 'graui'},
             {'token': 'piaque', 'lemma': 'lemma12', 'pos': 'pos12',
              'morph': 'Case=Case12|Numb=Numb12|Deg=Deg12|Mood=Mood12|Tense=Tense12|Voice=Voice12|Person=Person12',
              'treated': 'piaque'},
             {'token': 'lege', 'lemma': 'lemma13', 'pos': 'pos13',
              'morph': 'Case=Case13|Numb=Numb13|Deg=Deg13|Mood=Mood13|Tense=Tense13|Voice=Voice13|Person=Person13',
              'treated': 'lege'},
             {'token': 'noxiam', 'lemma': 'lemma14', 'pos': 'pos14',
              'morph': 'Case=Case14|Numb=Numb14|Deg=Deg14|Mood=Mood14|Tense=Tense14|Voice=Voice14|Person=Person14',
              'treated': 'noxiam'},
             {'token': 'lues', 'lemma': 'lemma15', 'pos': 'pos15',
              'morph': 'Case=Case15|Numb=Numb15|Deg=Deg15|Mood=Mood15|Tense=Tense15|Voice=Voice15|Person=Person15',
              'treated': 'lues'},
             {'token': '.', 'lemma': '.', 'pos': 'PUNC', 'morph': 'MORPH=empty', 'treated': '--IGN.--'},
             {'token': 'Licet', 'lemma': 'lemma16', 'pos': 'pos16',
              'morph': 'Case=Case16|Numb=Numb16|Deg=Deg16|Mood=Mood16|Tense=Tense16|Voice=Voice16|Person=Person16',
              'treated': 'licet'},
             {'token': 'querare', 'lemma': 'lemma17', 'pos': 'pos17',
              'morph': 'Case=Case17|Numb=Numb17|Deg=Deg17|Mood=Mood17|Tense=Tense17|Voice=Voice17|Person=Person17',
              'treated': 'querare'},
             {'token': ',', 'lemma': ',', 'pos': 'PUNC', 'morph': 'MORPH=empty', 'treated': '--IGN.--'},
             {'token': 'nec', 'lemma': 'lemma18', 'pos': 'pos18',
              'morph': 'Case=Case18|Numb=Numb18|Deg=Deg18|Mood=Mood18|Tense=Tense18|Voice=Voice18|Person=Person18',
              'treated': 'nec'},
             {'token': 'tibi', 'lemma': 'lemma19', 'pos': 'pos19',
              'morph': 'Case=Case19|Numb=Numb19|Deg=Deg19|Mood=Mood19|Tense=Tense19|Voice=Voice19|Person=Person19',
              'treated': 'tibi'},
             {'token': 'tener', 'lemma': 'lemma20', 'pos': 'pos20',
              'morph': 'Case=Case20|Numb=Numb20|Deg=Deg20|Mood=Mood20|Tense=Tense20|Voice=Voice20|Person=Person20',
              'treated': 'tener'},
             {'token': 'puer', 'lemma': 'lemma21', 'pos': 'pos21',
              'morph': 'Case=Case21|Numb=Numb21|Deg=Deg21|Mood=Mood21|Tense=Tense21|Voice=Voice21|Person=Person21',
              'treated': 'puer'}]
        )
